unit Unit1;

interface

uses
  Winapi.Windows, Winapi.Messages,
  System.SysUtils, System.Variants, System.Classes, System.ImageList, System.Actions, System.Generics.Collections,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.ExtCtrls, Vcl.ImgList, Vcl.ActnList,
  Vcl.ComCtrls, Vcl.ToolWin, uUndoManager;

type
  TForm1 = class(TForm)
    pnlMain: TPanel;
    dspShape: TShape;
    selShapeType: TRadioGroup;
    selShapeSize: TRadioGroup;
    dspChanges: TMemo;
    ToolBar1: TToolBar;
    ToolButton1: TToolButton;
    ToolButton2: TToolButton;
    ToolButton4: TToolButton;
    ToolButton3: TToolButton;
    ToolButton5: TToolButton;
    ActionList1: TActionList;
    actUndo: TAction;
    actRedo: TAction;
    actInitialize: TAction;
    actLockUndo: TAction;
    ImageList1: TImageList;
    ImageList2: TImageList;
    ImageList3: TImageList;
    procedure actRedoExecute(Sender: TObject);
    procedure actRedoUpdate(Sender: TObject);
    procedure actUndoExecute(Sender: TObject);
    procedure actUndoUpdate(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure selShapeTypeClick(Sender: TObject);
    procedure selShapeSizeClick(Sender: TObject);
  private
    FShapeSize: Integer;
    FShapeType: Integer;
    FUndoManager: TUndoManager;
    function GetUndoManager: TUndoManager;
    procedure LoadFromUndoStream(Stream: TStream);
    procedure SaveToUndoStream(Stream: TStream);
    procedure SetShapeSize(const Value: Integer);
    procedure SetShapeType(const Value: Integer);
    procedure ShapeSizeChanging(const Value: Integer);
    procedure ShapeTypeChanging(const Value: Integer);
  protected
    procedure BeginUndo;
    procedure CancelUndo;
    procedure EndUndo;
    procedure Redo;
    function RedoAvailable: Boolean;
    procedure Undo;
    function UndoAvailable: Boolean;
  public
    destructor Destroy; override;
    property ShapeSize: Integer read FShapeSize write SetShapeSize;
    property ShapeType: Integer read FShapeType write SetShapeType;
    property UndoManager: TUndoManager read GetUndoManager;
  end;

var
  Form1: TForm1;

implementation

const
  cShapeFactorHeight = 60;
  cShapeFactorWidth = 50;
  cHintHidePause = 10000;

// START resource string wizard section
resourcestring
  SInvalidShapeType = 'invalid shape type';
  SShapeTypeChangedTo = 'shape type changed to %d';
  SInvalidShapeSize = 'invalid shape size';
  SShapeSizeChangedTo = 'shape size changed to %d';

// END resource string wizard section


{$R *.dfm}

destructor TForm1.Destroy;
begin
  FreeAndNil(FUndoManager);
  inherited Destroy;
end;

procedure TForm1.actRedoExecute(Sender: TObject);
begin
  Redo;
end;

procedure TForm1.actRedoUpdate(Sender: TObject);
begin
  actRedo.Enabled := RedoAvailable;
end;

procedure TForm1.actUndoExecute(Sender: TObject);
begin
  Undo;
end;

procedure TForm1.actUndoUpdate(Sender: TObject);
begin
  actUndo.Enabled := UndoAvailable;
end;

procedure TForm1.BeginUndo;
begin
  UndoManager.BeginUndo;
end;

procedure TForm1.CancelUndo;
begin
  UndoManager.CancelUndo;
end;

procedure TForm1.EndUndo;
begin
  UndoManager.EndUndo;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  { Lass dem User genug Zeit, die langen Hints zu lesen... }
  Application.HintHidePause := cHintHidePause;

  { Damit auch wirklich eine nderung ausgelst wird. }
  FShapeType := -1;
  FShapeSize := -1;

  { RadioGroups werden synchronisiert }
  UndoManager.DisableUndo;
  try
    ShapeType := 0;
    ShapeSize := 1;
  finally
    UndoManager.EnableUndo;
  end;
  UndoManager.Clear;
  dspChanges.Lines.Clear;
end;

function TForm1.GetUndoManager: TUndoManager;
begin
  if FUndoManager = nil then begin
    FUndoManager := TUndoManager.Create;
    FUndoManager.OnLoadFromUndoStream := LoadFromUndoStream;
    FUndoManager.OnSaveToUndoStream := SaveToUndoStream;
  end;
  result := FUndoManager;
end;

procedure TForm1.LoadFromUndoStream(Stream: TStream);
var
  I: Integer;
  reader: TStreamReader;
begin
  reader := TStreamReader.Create(Stream);
  try
    ShapeType := reader.ReadLine.ToInteger;
    ShapeSize := reader.ReadLine.ToInteger;
    dspChanges.Lines.BeginUpdate;
    try
      dspChanges.Lines.Clear;
      for I := 1 to reader.ReadLine.ToInteger do begin
        dspChanges.Lines.Add(reader.ReadLine);
      end;
    finally
      dspChanges.Lines.EndUpdate;
    end;
  finally
    reader.Free;
  end;
end;

procedure TForm1.Redo;
begin
  UndoManager.Redo;
end;

function TForm1.RedoAvailable: Boolean;
begin
  Result := UndoManager.RedoAvailable;
end;

procedure TForm1.SaveToUndoStream(Stream: TStream);
var
  I: Integer;
  writer: TStreamWriter;
begin
  writer := TStreamWriter.Create(Stream);
  try
    writer.WriteLine(ShapeType);
    writer.WriteLine(ShapeSize);
    writer.WriteLine(dspChanges.Lines.Count);
    for I := 0 to dspChanges.Lines.Count - 1 do begin
      writer.WriteLine(dspChanges.Lines[I]);
    end;
  finally
    writer.Free;
  end;
end;

procedure TForm1.selShapeTypeClick(Sender: TObject);
begin
  ShapeType := selShapeType.ItemIndex;
end;

procedure TForm1.selShapeSizeClick(Sender: TObject);
begin
  ShapeSize := selShapeSize.ItemIndex;
end;

procedure TForm1.SetShapeSize(const Value: Integer);
begin
  if FShapeSize <> Value then
  begin
    try
      BeginUndo;
      try
        ShapeSizeChanging(Value);
        FShapeSize := Value;
        EndUndo;
      except
        CancelUndo;
        raise;
      end;
    finally
      selShapeSize.ItemIndex := FShapeSize;
    end;
  end;
end;

procedure TForm1.SetShapeType(const Value: Integer);
begin
  if FShapeType <> Value then
  begin
    try
      BeginUndo;
      try
        ShapeTypeChanging(Value);
        FShapeType := Value;
        EndUndo;
      except
        CancelUndo;
        raise;
      end;
    finally
      selShapeType.ItemIndex := FShapeType;
    end;
  end;
end;

procedure TForm1.ShapeSizeChanging(const Value: Integer);
begin
  if Value > 0 then begin
    dspShape.Width := cShapeFactorWidth*Value;
    dspShape.Height := cShapeFactorHeight*Value;
  end
  else begin
    raise ERangeError.Create(SInvalidShapeSize);
  end;
  dspChanges.Lines.Add(Format(SShapeSizeChangedTo, [Value]));
end;

procedure TForm1.ShapeTypeChanging(const Value: Integer);
begin
  case Value of
    0: dspShape.Shape := stRectangle;
    1: dspShape.Shape := stSquare;
    2: dspShape.Shape := stRoundRect;
    3: dspShape.Shape := stRoundSquare;
    4: dspShape.Shape := stEllipse;
    5: dspShape.Shape := stCircle;
  else
    raise ERangeError.Create(SInvalidShapeType);
  end;
  dspChanges.Lines.Add(Format(SShapeTypeChangedTo, [Value]));
end;

procedure TForm1.Undo;
begin
  UndoManager.Undo;
end;

function TForm1.UndoAvailable: Boolean;
begin
  Result := UndoManager.UndoAvailable;
end;

end.

